///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2003, 2004, 2005, Intel Corporation
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//      * Redistributions of source code must retain the above copyright notice,
//        this list of conditions and the following disclaimer.
//      * Redistributions in binary form must reproduce the above copyright
//        notice, this list of conditions and the following disclaimer in the
//        documentation and/or other materials provided with the distribution.
//      * Neither the name of the Intel Corporation nor the names of its
//        contributors may be used to endorse or promote products derived from
//        this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//  POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef LINUX
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#define APP_DESCRIPTOR int
#define PIPE_ERROR -1
#define PIPE_NAME "/var/tmp/mypipe"

#else // for Windows
#include <windows.h>
#include <stdio.h>

#define APP_DESCRIPTOR HANDLE
#define PIPE_ERROR INVALID_HANDLE_VALUE
extern void pWinError (char *function);
#define PIPE_NAME "\\\\.\\pipe\\mypipe"

#endif

#ifdef LINUX

int
CreateFullDuplexPipe (char *pipeName, APP_DESCRIPTOR * readFd,
                      APP_DESCRIPTOR * writeFd)
{
  int ret;

  // Create a named pipe in the Linux filesystem. 
  if ((ret = mkfifo (pipeName, S_IRWXU)) != 0)
    {
      // Don't worry if named pipe already exists
      if (errno != EEXIST)
        {
          perror ("CreatedFullDuplexPipe - mkfifo");
          return -1;
        }
    }

  // Open it for read and write.  Only a single descriptor is 
  // used/needed here for both the reader and writer, as the pipe is open
  // for both
  if ((*readFd = *writeFd = open (pipeName, O_RDWR)) < 0)
    {
      perror ("CreateFullDuplexPipe - open");
      return -1;
    }
  return 0;
}

size_t
WritePipe (int pipeFd, int numBytesToWrite, char *buf)
{

  size_t amount = write (pipeFd, buf, numBytesToWrite);
  if (amount == -1)
    {
      perror ("WritePipe - write");
      return -1;
    }
  return amount;
}

size_t
ReadPipe (int pipeFd, int numBytesToRead, char *buf)
{
  int amount = read (pipeFd, buf, numBytesToRead);
  if (amount == -1)
    {
      perror ("ReadPipe - read");
      return -1;
    }
  return amount;
}

int
ClosePipe (int readFd, int writeFd, char *pipeName)
{
  int ret;

  // Close the pipe file
  // Note read/write FDs are both the same in Linux
  ret = close (readFd);
  if (ret == -1)
    perror ("ClosePipe - close");

  // And delete the entry in the filesystem
  ret = unlink (pipeName);
  if (ret == -1)
    perror ("ClosePipe - unlink");

}

#else // Windows

int
CreateFullDuplexPipe (char *pipeName, APP_DESCRIPTOR * readFd,
                      APP_DESCRIPTOR * writeFd)
{
  // Create the named pipe first
  *readFd = CreateNamedPipe (pipeName, PIPE_ACCESS_DUPLEX,  // Open for read  and write
                             PIPE_TYPE_BYTE,    // Byte stream mode
                             PIPE_UNLIMITED_INSTANCES,  // Max # of instances
                             4096,  // Input buffer size
                             4096,  // Output buffer size
                             -1,    // Default timeout for connect
                             NULL); // Security attributes

  if (*readFd == INVALID_HANDLE_VALUE)
    {
      pWinError ("CreateNamedPipe");
      return -1;
    }
  // Open the pipe 
  *writeFd = CreateFile (pipeName,
                         GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (*writeFd == INVALID_HANDLE_VALUE)
    {
      pWinError ("CreateFile");
      return 1;
    }
  return 0;
}

int
ReadPipe (HANDLE readFd, int numBytesToRead, char *buf)
{
  DWORD bytes_read;
  BOOL ret = ReadFile (readFd,  // handle to pipe 
                       buf,     // buffer to receive data 
                       numBytesToRead,  // number of bytes to read 
                       &bytes_read, // number of bytes read 
                       NULL);   // not overlapped I/O 
  if (ret)
    return bytes_read;
  else
    {
      pWinError ("ReadFle");
      return -1;
    }
}

int
WritePipe (HANDLE writeFd, int numBytesToWrite, char *buf)
{
  DWORD bytes_written;

  BOOL ret = WriteFile (writeFd,
                        buf,
                        numBytesToWrite,
                        &bytes_written,
                        NULL);
  if (ret)
    return bytes_written;
  else
    {
      pWinError ("WriteFile");
      return -1;
    }
}

void
ClosePipe (HANDLE readFd, HANDLE writeFd, char *pipeName)
{
  if (!CloseHandle (writeFd))
    pWinError ("CloseHandle - writeFd");
  if (!CloseHandle (readFd))
    pWinError ("CloseHandle - readFd");
}

#endif

void
Pipe_demo (void)
{
  int runningTotal, amountWritten, amountRead, i, ret;
  int singleWriteSize = 25;
  int singleReadSize = 50;
  int bytesToWrite = 1000;
  int bytesToRead = 1000;
  char bufferIn[1000];
  char bufferOut[1000];
  APP_DESCRIPTOR writeFd, readFd;
  unsigned char goodData = 1;

  // This is a very demo simple using named pipes.  A pipe is opened for both
  // read and write.  Data is them stuffed into the pipe, and then read
  // from it.  All activity is taking place from the same process/thread,
  // which would not be the case in a "real" application using pipe for 
  // inter-process communication.

  printf ("Writing to and reading from a named pipe\n");

  // Use a buffer full of nines.
  memset (bufferIn, '9', 1000);

  // We'll demo a two way named pipe.
  ret = CreateFullDuplexPipe (PIPE_NAME, &readFd, &writeFd);
  if (ret == -1)
    return;

  // Write out 1K in 25-byte chunks.  Note that the byte counting below will not
  // handle some odd number of bytes.
  runningTotal = 0;
  printf ("Writing to pipe ");
  while (runningTotal < bytesToWrite)
    {

      amountWritten =
        WritePipe (writeFd, singleWriteSize, bufferIn + runningTotal);
      runningTotal += amountWritten;
      printf (".");
    }
  printf ("\nWrote %d bytes\n", runningTotal);

  // Now read the same number back from the pipe, but in 50 byte chunks
  // Normally, this would be done by another process after it opened the pipe.
  runningTotal = 0;
  printf ("Reading from pipe ");
  while (runningTotal < bytesToRead)
    {

      amountRead =
        ReadPipe (readFd, singleReadSize, bufferOut + runningTotal);
      runningTotal += amountRead;
      printf (".");
    }
  printf ("\nRead %d bytes\n", runningTotal);

  // Verify that all 9's were read
  for (i = 0; i < bytesToRead; i++)
    {
      //printf("%c", bufferOut[i]);
      if (bufferOut[i] != '9')
        {
          printf ("Bad data read from pipe!\n");
          goodData = 0;
          break;
        }
    }
  if (goodData)
    printf ("Data read from pipe was the same as data written to pipe\n");

  // Done with pipe; close and delete it
  ClosePipe (readFd, writeFd, PIPE_NAME);

}
