///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2003, 2004, 2005, Intel Corporation
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//      * Redistributions of source code must retain the above copyright notice,
//        this list of conditions and the following disclaimer.
//      * Redistributions in binary form must reproduce the above copyright
//        notice, this list of conditions and the following disclaimer in the
//        documentation and/or other materials provided with the distribution.
//      * Neither the name of the Intel Corporation nor the names of its
//        contributors may be used to endorse or promote products derived from
//        this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//  POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef LINUX

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <fcntl.h>

#else

#include <windows.h>
#include <stdio.h>

extern void pWinError (char *function);

#endif

// #defines common to both
#define FILE_NAME "mmap_file"
#define MEM_SIZE 1000


#ifdef LINUX

// Linux shared memory utilities

#define FILE_ID int

unsigned char *
AttachSharedMem (long memSize, char *fileName, FILE_ID memFileID)
{
  unsigned char *MappedFileMemPtr;

  // Map the file to the process' memory space
  // Note: file name not used - Windows only

  if ((MappedFileMemPtr =
       mmap (0, memSize, PROT_READ | PROT_WRITE, MAP_SHARED, memFileID,
             0)) == (unsigned char *) -1)
    {
      perror ("AttachedSharedMem - mmap");
      return (unsigned char *) 0;
    }
  else
    {
      printf ("File %s opened and mapped to address %X\n", FILE_NAME,
              MappedFileMemPtr);
      return MappedFileMemPtr;
    }
}

unsigned char *
InitSharedMem (long shMemSize, char *fileName, FILE_ID * memFileID)
{

  // Called to first open, size and map the file to a process' memory.
  // Open the specified named file for read/write.  Create it if it doesn't
  // exist, truncate it to 0 if it does
  if ((*memFileID = open (fileName, O_RDWR | O_CREAT | O_TRUNC)) == -1)
    {
      perror ("InitSharedMem - open");
      return (unsigned char *) 0;
    }
  else
    {
      // Open up space in the file
      if (lseek (*memFileID, shMemSize - 1, SEEK_SET) == -1)
        {
          perror ("InitSharedMem - lseek");
          return (unsigned char *) 0;
        }
      if (write (*memFileID, "", 1) != 1)
        {
          perror ("InitSharedMem - write");
          return (unsigned char *) 0;
        }
      return AttachSharedMem (shMemSize, "NoNameNeeded", *memFileID);
    }
}

void
DeleteSharedMem (void *memAddr, FILE_ID memFileID)
{
  if (munmap (memAddr, MEM_SIZE) == -1)
    perror ("DeleteSharedMem - munmap");
}

#else // Windows

#define FILE_ID HANDLE

unsigned char *
InitSharedMem (long ShMemSize, char *fileName, HANDLE * mapHandle)
{
  // Open a new file for read and write.  Then, initialize an area of shared memory 
  // of the requested size (in bytes) mapped to the file
  // Return a pointer to the start of memory, NULL if a problem.
  LPCTSTR buf;
  HANDLE fileHandle;

  fileHandle = CreateFile (TEXT (FILE_NAME),    // file to create
                           GENERIC_WRITE | GENERIC_READ,    // open for read/write
                           FILE_SHARE_WRITE | FILE_SHARE_READ,  // share for read/write
                           NULL,    // default security
                           CREATE_ALWAYS,   // overwrite existing
                           FILE_ATTRIBUTE_NORMAL,   // normal file
                           NULL);   // no attr. template

  if (fileHandle == INVALID_HANDLE_VALUE)
    {
      pWinError ("CreateFile");
      return NULL;
    }

  *mapHandle = CreateFileMapping (fileHandle,   // use file opened above
                                  NULL, // default security 
                                  PAGE_READWRITE,   // read/write access
                                  0,    // max. object size 
                                  ShMemSize,    // buffer size  
                                  fileName);    // name of mapping object

  if (*mapHandle == NULL || *mapHandle == INVALID_HANDLE_VALUE)
    {
      pWinError ("CreateFileMapping");
      return NULL;
    }
  buf = (LPTSTR) MapViewOfFile (*mapHandle, // handle to map object
                                FILE_MAP_ALL_ACCESS,    // read/write permission
                                0,  // possible offset in file   
                                0,  // possible offset in file     
                                ShMemSize); // number of bytes of file mapping to map to view

  if (buf == NULL)
    {
      pWinError ("MapViewOfFile");
      return NULL;
    }
  return (unsigned char *) buf;
}

unsigned char *
AttachSharedMem (long ShMemSize, char *fileName, FILE_ID memFileID)
{
  HANDLE fileHandle;
  LPCTSTR buf;

  // Mem mapped file already exists, just need access to it
  // Note: memID not used - Linux only
  fileHandle = OpenFileMapping (FILE_MAP_ALL_ACCESS,    // read/write access
                                FALSE,  // do not inherit the name
                                fileName);  // name of mapping object 

  if (fileHandle == NULL)
    {
      pWinError ("OpenFileMapping");
      return NULL;
    }

  buf = MapViewOfFile (fileHandle,  // handle to mapping object
                       FILE_MAP_ALL_ACCESS, // read/write permission
                       0, 0, ShMemSize);


  if (buf == NULL)
    {
      pWinError ("MapViewOfFile");
      return NULL;
    }
  return (unsigned char *) buf;
}

void
DeleteSharedMem (const char *memAddr, HANDLE fileHandle)
{
  if (UnmapViewOfFile (memAddr) == 0)
    {
      pWinError ("UnmapViewOfFile");
    }

  if (fileHandle == INVALID_HANDLE_VALUE)
    if (CloseHandle (fileHandle) == 0)
      {
        pWinError ("CloseHandle");
      }

}

#endif

void
SharedMem_demo (void)
{
  int i;
  unsigned char *FirstSharedMemPtr;
  unsigned char *SecondSharedMemPtr;
  FILE_ID memFileID;

  // Demo that opens a file, then maps it to a processe' memory.  A second view of the
  // file is then opened.  data is written into one view, then read and verified from the other.
  // The file left at then end of the test shoudl contain 1000 9's.

  // Note: normally this would be done from two different processes, but the principles 
  // here are the same.

  printf
    ("\nCreating a 1K memory-mapped file and getting two views of it for this process\n");

  FirstSharedMemPtr = InitSharedMem (MEM_SIZE, FILE_NAME, &memFileID);
  SecondSharedMemPtr = AttachSharedMem (MEM_SIZE, FILE_NAME, memFileID);

  // Write some data into the memory segment
  printf ("Writing data to memory and file %s\n", FILE_NAME);
  for (i = 0; i < MEM_SIZE; i++)
    *(FirstSharedMemPtr + i) = '9';

  // Verify what's there using the other view of the memory
  printf ("Verifying data in memory and file %s\n", FILE_NAME);
  for (i = 0; i < MEM_SIZE; i++)
    {
      if (*(SecondSharedMemPtr + i) != '9')
        printf ("Bad data found in shared memory segment\n");
    }

  // Clean up shared memory
  DeleteSharedMem (FirstSharedMemPtr, memFileID);
  DeleteSharedMem (SecondSharedMemPtr, memFileID);
}
