///////////////////////////////////////////////////////////////////////////////
//  Copyright (c) 2003, 2004, 2005, Intel Corporation
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//      * Redistributions of source code must retain the above copyright notice,
//        this list of conditions and the following disclaimer.
//      * Redistributions in binary form must reproduce the above copyright
//        notice, this list of conditions and the following disclaimer in the
//        documentation and/or other materials provided with the distribution.
//      * Neither the name of the Intel Corporation nor the names of its
//        contributors may be used to endorse or promote products derived from
//        this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS AS IS
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
//  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
//  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//  POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef LINUX
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <sys/timeb.h>
#include <curl/curl.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#else // for Windows

#include <windows.h>
#include <stdio.h>
#include <time.h>

// An equivalent to Linux perror()
void
pWinError (char *function)
{
  LPVOID MsgBuf;
  DWORD err = GetLastError ();
  FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
                 FORMAT_MESSAGE_FROM_SYSTEM,
                 NULL,
                 err,
                 MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                 (LPTSTR) & MsgBuf, 0, NULL);
  printf ("Error in %s: (%d) %s\n", function, err, MsgBuf);
  LocalFree (MsgBuf);
}

#endif


// References to things outside this file
extern char *LocalHostIPAddr (char *, size_t);
extern void InitTimer (long);
extern void enable_int_handlers (void);
extern int errno;
extern unsigned char *InitSharedMem (long ShMemSize);
extern void DeleteSharedMem (void);
extern void ChangeSharedMemMode (char *mode);
extern void SharedMem_demo (void);
extern void Pipe_demo (void);

// A global for interrupt handler test
extern int GapplicationRunning;


#ifndef LINUX
#endif

// Several of the simple utilities are here.  Most are in separate files

// Sleep()/usleep() compatibility
#ifdef LINUX

void
app_sleep (unsigned long millisecs)
{
  usleep (millisecs * 1000);
}

#else

void
app_sleep (unsigned long millisecs)
{
  Sleep (millisecs);
}

#endif


void
BuildTimestamp (char *timestamp)
{

#ifdef LINUX

  struct timeb tstruct;
  char *timeline;
  ftime (&tstruct);
  timeline = ctime (&(tstruct.time));
  sprintf (timestamp, "%.19s.%hu", timeline, tstruct.millitm);

#else

  char timestr[128];
  SYSTEMTIME systemTime;
  GetLocalTime (&systemTime);

  sprintf (timestr,
           "%d/%02d/%02d %02d:%02d:%02d.%03d",
           systemTime.wYear, systemTime.wMonth, systemTime.wDay,
           systemTime.wHour, systemTime.wMinute, systemTime.wSecond,
           systemTime.wMilliseconds);

  sprintf (timestamp, "%s", timestr);

#endif
}


// Return the ID of the current thread
long
GetCurThreadID ()
{

#ifdef LINUX

  pthread_t this_thread = pthread_self ();
  return ((long) this_thread);

#else

  return (GetCurrentThreadId ());

#endif
}

int
main (void)
{
  char ipaddrbuf[32];
  char ts_buf[80];
  long msTimer = 1500;
  char c;

  printf
    ("Show some areas where Windows and Linux system calls differ:\n\nSleep for 2 seconds...\n");
  app_sleep (2000);
  printf ("Done\n\n");
  BuildTimestamp (ts_buf);
  printf ("Timestamp for NOW is %s\n\n", ts_buf);
  printf ("This is thread number %d in this process.\n\n", GetCurThreadID ());
  printf ("Local host IP address is %s\n\n",
          LocalHostIPAddr (ipaddrbuf, sizeof (ipaddrbuf)));

  Pipe_demo ();
  SharedMem_demo ();

  printf ("\nSetting a timer to expire in %d milliseconds\n", msTimer);
  InitTimer (msTimer);
  enable_int_handlers ();
  while (GapplicationRunning)
    {
      printf ("Waiting for a Ctrl-C interrupt...\n");
      c = getchar ();
    }

  return (0);
}
